
# Local library
from .. import lib
import pyblish.plugin
import pyblish.logic
from nose.tools import (
    with_setup,
    assert_equals,
    assert_raises,
    assert_true,
    assert_false
)


@with_setup(lib.setup_empty, lib.teardown)
def test_di():
    """Dependency injection works fine"""

    _disk = dict()

    # Plugins
    class SelectInstance(pyblish.api.Selector):
        def process(self, context):
            self.log.info("Test")
            for name in ("MyInstanceA", "MyInstanceB"):
                instance = context.create_instance(name)
                instance.set_data("family", "myFamily")
                instance.set_data("value", "123")

            print("Instance: %s" % instance)

    class ValidateInstance(pyblish.api.Validator):
        def process(self, instance):
            assert instance.data("family") == "myFamily", "Wrong family"

    class ExtractInstanceX(pyblish.api.Extractor):
        def process(self, context, instance, user, time):
            self.log.warning("Filling up disk..")
            _disk[instance.name] = "%s - %s: %s" % (
                time(), user, instance.data("value"))

    for plugin in (SelectInstance, ValidateInstance, ExtractInstanceX):
        pyblish.api.register_plugin(plugin)

    context = pyblish.api.Context()

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=pyblish.api.discover(),
            context=context):
        assert result["error"] is None, result["error"]

    assert "MyInstanceB" in _disk


@with_setup(lib.setup_empty, lib.teardown)
def test_init():
    """__init__ is triggered once per process"""

    count = {"#": 0}

    class HappensOnce(pyblish.api.Selector):
        def __init__(self):
            count["#"] += 1

        def process(self, context):
            for name in ("Smurfette", "Passive-aggressive smurf"):
                instance = context.create_instance(name)
                instance.set_data("family", "smurfFamily")

    class HappensTwice(pyblish.api.Validator):
        def __init__(self):
            count["#"] += 10

        def process(self, instance):
            pass

    for plugin in (HappensOnce,
                   HappensTwice):
        pyblish.api.register_plugin(plugin)

    list(pyblish.logic.process(
        func=pyblish.plugin.process,
        plugins=pyblish.api.discover(),
        context=pyblish.api.Context()))

    assert_equals(count["#"], 21)


@with_setup(lib.setup_empty, lib.teardown)
def test_occurence():
    """Test when and how often plug-ins process"""

    count = {"#": 0}

    class HappensOnce1(pyblish.api.Selector):
        def process(self, context):
            count["#"] += 1
            for name in ("Smurfette", "Passive-aggressive smurf"):
                instance = context.create_instance(name)
                instance.set_data("family", "smurfFamily")

    class HappensOnce2(pyblish.api.Validator):
        def process(self):
            count["#"] += 1

    class DoesNotHappen1(pyblish.api.Validator):
        """Doesn't run

        It's requesting to be triggered only in the presence
        of an instance of a family that isn't present.

        """

        families = ["unsupportedFamily"]

        def process(self):
            self.log.critical(str(self))
            count["#"] += 1

    class DoesNotHappen2(pyblish.api.Validator):
        """Doesn't run.

        Asking for `instance`, and only supporting an unavailable
        family prevents this plug-in from running

        """

        families = ["unsupportedFamily"]

        def process(self, instance):
            self.log.critical(str(self))
            count["#"] += 1

    class HappensEveryInstance(pyblish.api.Validator):
        def process(self, instance):
            count["#"] += 1

    for plugin in (HappensOnce1,
                   HappensOnce2,
                   DoesNotHappen1,
                   DoesNotHappen2,
                   HappensEveryInstance):
        pyblish.api.register_plugin(plugin)

    context = pyblish.api.Context()

    list(pyblish.logic.process(
        func=pyblish.plugin.process,
        plugins=pyblish.api.discover(),
        context=context))

    assert_equals(count["#"], 4)


@with_setup(lib.setup_empty, lib.teardown)
def test_no_instances():
    """Run .process at least once per plug-in"""

    count = {"#": 0}

    class Extract(pyblish.api.Extractor):
        def process(self, context):
            count["#"] += 1

    class Extract2(pyblish.api.Extractor):
        """This doesn't run

        Because it asks for instance, and there aren't any instances

        """

        def process(self, context, instance):
            assert instance is None
            count["#"] += 1

    class Conform(pyblish.api.Conformer):
        def process(self, context):
            count["#"] += 1

    for plugin in (Extract, Extract2, Conform):
        pyblish.api.register_plugin(plugin)

    context = pyblish.api.Context()

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=pyblish.api.discover(),
            context=context):
        pass

    assert_equals(count["#"], 2)


def test_unavailable_service():
    """Asking for unavailable service throws exception"""

    provider = pyblish.plugin.Provider()

    def func(arg1, arg2):
        return True

    provider.inject("arg1", lambda: True)
    assert_raises(KeyError, provider.invoke, func)


def test_unavailable_service_logic():
    """Asking for unavailable service ..?"""

    class SelectUnavailable(pyblish.api.Selector):
        def process(self, unavailable):
            print("HHOH")
            self.log.critical("Test")

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=[SelectUnavailable],
            context=pyblish.api.Context()):
        assert_true(isinstance(result["error"], KeyError))


@with_setup(lib.setup_empty, lib.teardown)
def test_test_failure():
    """Failing the test yields an exception"""

    triggered = list()

    class ValidateFailure(pyblish.api.Validator):
        def process(self, context):
            triggered.append(self)
            assert False

    class ExtractFailure(pyblish.api.Extractor):
        def process(self, context):
            triggered.append(self)
            pass

    pyblish.api.register_plugin(ValidateFailure)
    pyblish.api.register_plugin(ExtractFailure)

    context = pyblish.api.Context()

    results = list(pyblish.logic.process(
        func=pyblish.plugin.process,
        plugins=pyblish.api.discover(),
        context=context))

    assert_equals(len(triggered), 1)
    assert_equals(type(triggered[0]).id, ValidateFailure.id)
    assert_true(isinstance(results[-1], Exception))


@with_setup(lib.setup_empty, lib.teardown)
def test_when_to_trigger_process():
    """process() should be triggered whenever `context` is requested"""

    _data = {"error": False}

    class SelectInstance(pyblish.api.Selector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.set_data("family", "compatibleFamily")

    class IncompatibleValidator(pyblish.api.Validator):
        families = ["incompatibleFamily"]

        def process(self, instance):
            print("Instance is: %s" % instance)
            _data["error"] = True
            assert False, "I should not have been run"

    class CompatibleValiator(pyblish.api.Validator):
        families = ["compatibleFamily"]

        def process(self, instance):
            assert True

    for plugin in (SelectInstance, IncompatibleValidator, CompatibleValiator):
        pyblish.api.register_plugin(plugin)

    context = pyblish.api.Context()
    list(pyblish.logic.process(
        func=pyblish.plugin.process,
        plugins=pyblish.api.discover(),
        context=context))

    assert_equals(_data["error"], False)


def test_default_services():
    """Default services are operational"""

    services = ["user", "time"]
    for service in services:
        assert_true(service in pyblish.api.registered_services())

    # user is passed by value, as it does not change at run-time
    user = pyblish.api.registered_services()["user"]
    assert_false(hasattr(user, "__call__"))

    # time is passed by reference, and is callable, as it changes
    time = pyblish.api.registered_services()["time"]
    time()


def test_asset():
    """Assets with DI works well"""

    count = {"#": 0}

    class SelectCharacters(pyblish.api.Selector):
        """Called once"""
        def process(self, context):
            for name in ("A", "B"):
                context.create_asset(name, family="myFamily")
            count["#"] += 1

    class ValidateColor(pyblish.api.Validator):
        """Called twice"""
        families = ["myFamily"]

        def process(self, asset):
            count["#"] += 1

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=[SelectCharacters, ValidateColor],
            context=pyblish.api.Context()):
        print(result)

    assert_equals(count["#"], 3)


@with_setup(lib.setup_empty, lib.teardown)
def test_di_testing():
    """DI simplifies testing"""

    instances = list()

    class SelectCharacters(pyblish.api.Validator):
        def process(self, context, host):
            for char in host.ls("*_char"):
                instance = context.create_instance(char, family="character")
                instance.add(host.listRelatives(char))
                instances.append(instance.name)

    class HostMock(object):
        def ls(self, query):
            return ["bobby_char", "rocket_char"]

        def listRelatives(self, node):
            if node == "bobby_char":
                return ["arm", "leg"]
            if node == "rocket_char":
                return ["propeller"]
            return []

    pyblish.api.register_service("host", HostMock())

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=[SelectCharacters],
            context=pyblish.api.Context()):
        assert_equals(result["error"], None)

    assert_equals(len(instances), 2)
    assert_equals(instances, ["bobby_char", "rocket_char"])



import os
import contextlib

# Local library
from . import lib

from pyblish import api, logic, plugin, util

from nose.tools import (
    with_setup,
    assert_equals,
)


@contextlib.contextmanager
def no_guis():
    os.environ.pop("PYBLISHGUI", None)
    for gui in logic.registered_guis():
        logic.deregister_gui(gui)

    yield


@with_setup(lib.setup, lib.teardown)
def test_iterator():
    """Iterator skips inactive plug-ins and instances"""

    count = {"#": 0}

    class MyCollector(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            inactive = context.create_instance("Inactive")
            active = context.create_instance("Active")

            inactive.data["publish"] = False
            active.data["publish"] = True

            count["#"] += 1

    class MyValidatorA(api.InstancePlugin):
        order = api.ValidatorOrder
        active = False

        def process(self, instance):
            count["#"] += 10

    class MyValidatorB(api.InstancePlugin):
        order = api.ValidatorOrder

        def process(self, instance):
            count["#"] += 100

    context = api.Context()
    plugins = [MyCollector, MyValidatorA, MyValidatorB]

    assert count["#"] == 0, count

    for Plugin, instance in logic.Iterator(plugins, context):
        assert instance.name != "Inactive" if instance else True
        assert Plugin.__name__ != "MyValidatorA"

        plugin.process(Plugin, context, instance)

    # Collector runs once, one Validator runs once
    assert count["#"] == 101, count


def test_register_gui():
    """Registering at run-time takes precedence over those from environment"""

    with no_guis():
        os.environ["PYBLISHGUI"] = "second,third"
        logic.register_gui("first")

        print(logic.registered_guis())
        assert logic.registered_guis() == ["first", "second", "third"]

    with no_guis():
        os.environ["PYBLISH_GUI"] = "second,third"
        logic.register_gui("first")

        print(logic.registered_guis())
        assert logic.registered_guis() == ["first", "second", "third"]


@with_setup(lib.setup_empty, lib.teardown)
def test_subset_match():
    """Plugin.match = api.Subset works as expected"""

    count = {"#": 0}

    class MyPlugin(api.InstancePlugin):
        families = ["a", "b"]
        match = api.Subset

        def process(self, instance):
            count["#"] += 1

    context = api.Context()

    context.create_instance("not_included_1", families=["a"])
    context.create_instance("not_included_1", families=["x"])
    context.create_instance("included_1", families=["a", "b"])
    context.create_instance("included_2", families=["a", "b", "c"])

    util.publish(context, plugins=[MyPlugin])

    assert_equals(count["#"], 2)

    instances = logic.instances_by_plugin(context, MyPlugin)
    assert_equals(list(i.name for i in instances),
                  ["included_1", "included_2"])


def test_subset_exact():
    """Plugin.match = api.Exact works as expected"""

    count = {"#": 0}

    class MyPlugin(api.InstancePlugin):
        families = ["a", "b"]
        match = api.Exact

        def process(self, instance):
            count["#"] += 1

    context = api.Context()

    context.create_instance("not_included_1", families=["a"])
    context.create_instance("not_included_1", families=["x"])
    context.create_instance("not_included_3", families=["a", "b", "c"])
    instance = context.create_instance("included_1", families=["a", "b"])

    # Discard the solo-family member, which defaults to `default`.
    #
    # When using multiple families, it is common not to bother modifying
    # `family`, and in the future this member needn't be there at all and
    # may/should be removed. But till then, for complete clarity, it might
    # be worth removing this explicitly during the creation of instances
    # if instead choosing to use the `families` key.
    instance.data.pop("family")

    util.publish(context, plugins=[MyPlugin])

    assert_equals(count["#"], 1)

    instances = logic.instances_by_plugin(context, MyPlugin)
    assert_equals(list(i.name for i in instances), ["included_1"])


def test_plugins_by_families():
    """The right plug-ins are returned from plugins_by_families"""

    class ClassA(api.Collector):
        families = ["a"]

    class ClassB(api.Collector):
        families = ["b"]

    class ClassC(api.Collector):
        families = ["c"]

    class ClassD(api.Collector):
        families = ["a", "b"]
        match = api.Intersection

    class ClassE(api.Collector):
        families = ["a", "b"]
        match = api.Subset

    class ClassF(api.Collector):
        families = ["a", "b"]
        match = api.Exact

    assert logic.plugins_by_families(
        [ClassA, ClassB, ClassC], ["a", "z"]) == [ClassA]

    assert logic.plugins_by_families(
        [ClassD, ClassE, ClassF], ["a"]) == [ClassD]

    assert logic.plugins_by_families(
        [ClassD, ClassE, ClassF], ["a", "b"]) == [ClassD, ClassE, ClassF]

    assert logic.plugins_by_families(
        [ClassD, ClassE, ClassF], ["a", "b", "c"]) == [ClassD, ClassE]



"""Expose common functionality

Use this in plugins, integrations and extensions of Pyblish.
This part of the library is what will change the least and
attempt to maintain backwards- and forwards-compatibility.

This way, we as developers are free to refactor the library
without breaking any of your tools.

.. note:: Contributors, don't use this in any other module internal
    to Pyblish or a cyclic dependency will be created. This is only
    to be used by end-users of the library and from
    integrations/extensions of Pyblish.

"""

from __future__ import absolute_import

from . import version
import getpass
import os

from .plugin import (
    Context,
    Instance,

    Action,
    Category,
    Separator,

    # Matching algorithms
    Subset,
    Intersection,
    Exact,

    Asset,
    Plugin,
    Validator,
    Extractor,
    Integrator,
    Collector,
    discover,

    ContextPlugin,
    InstancePlugin,

    CollectorOrder,
    ValidatorOrder,
    ExtractorOrder,
    IntegratorOrder,

    register_host,
    registered_hosts,
    deregister_host,
    deregister_all_hosts,

    current_target,
    register_target,
    registered_targets,
    deregister_target,
    deregister_all_targets,

    register_plugin,
    deregister_plugin,
    deregister_all_plugins,
    registered_plugins,

    plugin_paths,
    register_plugin_path,
    deregister_plugin_path,
    deregister_all_paths,

    register_service,
    deregister_service,
    deregister_all_services,
    registered_services,

    register_callback,
    deregister_callback,
    deregister_all_callbacks,
    registered_callbacks,

    sort as sort_plugins,

    registered_paths,
    environment_paths,
    current_host,
)

from .lib import (
    log,
    time as __time,
    emit,
    main_package_path as __main_package_path
)

from .logic import (
    plugins_by_family,
    plugins_by_host,
    plugins_by_instance,
    plugins_by_targets,
    instances_by_plugin,
    register_test,
    deregister_test,
    registered_test,

    register_gui,
    registered_guis,
    deregister_gui,

    default_test as __default_test,
)

from .error import (
    PyblishError,
    SelectionError,
    ValidationError,
    ExtractionError,
    ConformError,
    NoInstancesError
)

from .compat import (
    deregister_all,
    sort,
    Selector,
    Conformer,
    format_filename,
)


def __init__():
    """Initialise Pyblish

    This function registered default services,
    hosts and tests. It is idempotent and thread-safe.

    """

    # Register default services
    register_service("time", __time)
    register_service("user", getpass.getuser())
    register_service("context", None)
    register_service("instance", None)

    # Register default host
    register_host("python")

    # Register hosts from environment "PYBLISHHOSTS"
    for host in os.environ.get("PYBLISH_HOSTS", "").split(os.pathsep):
        if host:
            register_host(host)

    # Register default path
    register_plugin_path(os.path.join(__main_package_path(), "plugins"))

    # Register default test
    register_test(__default_test)


__init__()


__all__ = [
    # Base objects
    "Context",
    "Instance",
    "Asset",

    # Matching algorithms
    "Subset",
    "Intersection",
    "Exact",

    "Plugin",
    "Action",
    "Category",
    "Separator",

    # SVEC plug-ins
    "Collector",
    "Selector",
    "Validator",
    "Extractor",
    "Conformer",
    "Integrator",

    "ContextPlugin",
    "InstancePlugin",

    "CollectorOrder",
    "ValidatorOrder",
    "ExtractorOrder",
    "IntegratorOrder",

    # Plug-in utilities
    "discover",

    "plugin_paths",
    "registered_paths",
    "environment_paths",

    "register_host",
    "registered_hosts",
    "deregister_host",
    "deregister_all_hosts",

    "register_plugin",
    "deregister_plugin",
    "deregister_all_plugins",
    "registered_plugins",

    "register_service",
    "deregister_service",
    "deregister_all_services",
    "registered_services",

    "register_callback",
    "deregister_callback",
    "deregister_all_callbacks",
    "registered_callbacks",

    "register_plugin_path",
    "deregister_plugin_path",
    "deregister_all_paths",

    "register_test",
    "deregister_test",
    "registered_test",

    "register_gui",
    "registered_guis",
    "deregister_gui",

    "plugins_by_family",
    "plugins_by_host",
    "plugins_by_instance",
    "instances_by_plugin",

    "current_target",
    "register_target",
    "registered_targets",
    "deregister_target",
    "deregister_all_targets",

    "sort_plugins",
    "format_filename",
    "current_host",
    "sort_plugins",

    "version",

    # Utilities
    "log",
    "emit",

    # Exceptions
    "PyblishError",
    "SelectionError",
    "ValidationError",
    "ExtractionError",
    "ConformError",
    "NoInstancesError",

    # Compatibility
    "deregister_all",
    "sort",
]



import os
import logging

from pyblish.vendor import mock
import pyblish.api
import pyblish.util
import pyblish.plugin
from nose.tools import (
    with_setup,
    assert_true,
    assert_equals,
    assert_raises,
    raises,
)

from . import lib


@with_setup(lib.setup_empty, lib.teardown)
def test_unique_id():
    """Plug-ins and instances have an id"""

    class MyPlugin(pyblish.plugin.Collector):
        pass

    class MyAction(pyblish.plugin.Action):
        pass

    assert_true(hasattr(MyPlugin, "id"))

    instance = pyblish.plugin.Instance("MyInstance")
    assert_true(hasattr(instance, "id"))

    # IDs are persistent
    assert_equals(instance.id, instance.id)
    assert_equals(MyAction.id, MyAction.id)
    assert_equals(MyPlugin.id, MyPlugin.id)

    context = pyblish.plugin.Context()
    assert_equals(context.id, context.id)

    # Even across discover()'s
    # Due to the fact that an ID is generated on module
    # load, which only happens once per process unless
    # module is forcefully reloaded by the user.
    pyblish.api.register_plugin(MyPlugin)
    plugins = list(p for p in pyblish.api.discover()
                   if p.id == MyPlugin.id)
    assert len(plugins) == 1, plugins
    assert plugins[0].__name__ == MyPlugin.__name__


def test_context_from_instance():
    """Instances provide access to their parent context"""

    context = pyblish.plugin.Context()
    instance = context.create_instance("MyInstance")
    assert_equals(context, instance.context)


def test_legacy():
    """Legacy is determined by existing process_* methods"""
    class LegacyPlugin(pyblish.plugin.Collector):
        def process_context(self, context):
            pass

    class NotLegacyPlugin(pyblish.plugin.Collector):
        def process(self, context):
            pass

    assert_true(hasattr(LegacyPlugin, "__pre11__"))
    assert_equals(LegacyPlugin.__pre11__, True)
    assert_true(hasattr(NotLegacyPlugin, "__pre11__"))
    assert_equals(NotLegacyPlugin.__pre11__, False)


def test_asset():
    """Using asset over instance works fine"""
    context = pyblish.plugin.Context()

    asseta = context.create_asset("MyAssetA", family="myFamily")
    assetb = context.create_asset("MyAssetB", family="myFamily")

    assert_true(asseta in context)
    assert_true(assetb in context)


@with_setup(lib.setup_empty, lib.teardown)
def test_import_mechanism_duplication():
    """Plug-ins don't linger after a second discovery

    E.g. when changing the name of a plug-in and then rediscover
    the previous plug-ins is still around.

    """

    with lib.tempdir() as temp:
        print("Writing temporarily to: %s" % temp)
        module = os.path.join(temp, "selector.py")
        pyblish.api.register_plugin_path(temp)

        with open(module, "w") as f:
            f.write("""
import pyblish.api

class MySelector(pyblish.api.Selector):
    pass
""")

        with open(module) as f:
            print("File contents after first write:")
            print(f.read())

        # MySelector should be accessible by now
        plugins = [p.__name__ for p in pyblish.api.discover()]

        assert "MySelector" in plugins, plugins
        assert "MyOtherSelector" not in plugins, plugins

        # Remove module, and it's .pyc equivalent
        [os.remove(os.path.join(temp, fname))
         for fname in os.listdir(temp)]

        with open(module, "w") as f:
            f.write("""
import pyblish.api

class MyOtherSelector(pyblish.api.Selector):
    pass
""")

        with open(module) as f:
            print("File contents after second write:")
            print(f.read())

        # MySelector should be gone in favour of MyOtherSelector
        plugins = [p.__name__ for p in pyblish.api.discover()]

        assert "MyOtherSelector" in plugins, plugins
        assert "MySelector" not in plugins, plugins


@raises(TypeError)
@with_setup(lib.setup_empty, lib.teardown)
def test_register_unsupported_hosts():
    """Cannot register a unsupported plug-in in an unsupported host"""

    class Unsupported(pyblish.api.Plugin):
        hosts = ["unsupported"]

    pyblish.api.register_plugin(Unsupported)


@raises(TypeError)
@with_setup(lib.setup_empty, lib.teardown)
def test_register_unsupported_version():
    """Cannot register a plug-in of an unsupported version"""

    class Unsupported(pyblish.api.Plugin):
        requires = (999, 999, 999)

    pyblish.api.register_plugin(Unsupported)


@raises(TypeError)
@with_setup(lib.setup_empty, lib.teardown)
def test_register_malformed():
    """Cannot register a malformed plug-in"""

    class Unsupported(pyblish.api.Plugin):
        families = True
        hosts = None

    pyblish.api.register_plugin(Unsupported)


@with_setup(lib.setup_empty, lib.teardown)
def test_temporarily_disabled_plugins():
    """Plug-ins as files starting with an underscore are hidden"""

    discoverable = """
import pyblish.api

class Discoverable(pyblish.api.Plugin):
    pass
"""

    notdiscoverable = """
import pyblish.api

class NotDiscoverable(pyblish.api.Plugin):
    pass
"""

    with lib.tempdir() as d:
        pyblish.api.register_plugin_path(d)

        with open(os.path.join(d, "discoverable.py"), "w") as f:
            f.write(discoverable)

        with open(os.path.join(d, "_undiscoverable.py"), "w") as f:
            f.write(notdiscoverable)

        plugins = [p.__name__ for p in pyblish.api.discover()]
        assert "Discoverable" in plugins
        assert "NotDiscoverable" not in plugins


@with_setup(lib.setup_empty, lib.teardown)
def test_repair_context_backwardscompat():
    """Plug-ins with repair-context are reprogrammed appropriately"""

    class ValidateInstances(pyblish.api.Validator):
        def repair_context(self, context):
            pass

    assert hasattr(ValidateInstances, "repair")
    assert not hasattr(ValidateInstances, "repair_context")


@with_setup(lib.setup_empty, lib.teardown)
def test_unique_logger():
    """A unique logger is applied to every plug-in"""

    count = {"#": 0}

    class MyPlugin(pyblish.api.Plugin):
        def process(self, context):
            self.log.debug("Hello world")
            count["#"] += 1

    pyblish.api.register_plugin(MyPlugin)

    context = pyblish.util.publish()

    assert_equals(count["#"], 1)
    print(context.data("results"))

    results = context.data("results")[0]
    records = results["records"]
    hello_world = records[0]
    assert_equals(hello_world.msg, "Hello world")

    pyblish.api.deregister_plugin(MyPlugin)


@with_setup(lib.setup_empty, lib.teardown)
def test_current_host():
    """pyblish.api.current_host works"""
    pyblish.plugin.register_host("myhost")
    assert_equals(pyblish.plugin.current_host(), "myhost")

    assert_raises(Exception, pyblish.plugin.deregister_host, "notExist")


@with_setup(lib.setup_empty, lib.teardown)
def test_register_host():
    """Registering and deregistering hosts works fine"""
    pyblish.plugin.register_host("myhost")
    assert "myhost" in pyblish.plugin.registered_hosts()
    pyblish.plugin.deregister_host("myhost")
    assert "myhost" not in pyblish.plugin.registered_hosts()


@with_setup(lib.setup_empty, lib.teardown)
def test_current_target():
    """pyblish.api.current_target works"""
    pyblish.plugin.register_target("mytarget")
    assert_equals(pyblish.plugin.current_target(), "mytarget")

    assert_raises(Exception, pyblish.plugin.deregister_target, "notExist")


@with_setup(lib.setup_empty, lib.teardown)
def test_current_target_latest():
    """pyblish.api.current_target works"""
    pyblish.plugin.deregister_all_targets()
    pyblish.plugin.register_target("mytarget1")
    pyblish.plugin.register_target("mytarget2")
    assert_equals(pyblish.plugin.current_target(), "mytarget2")

    pyblish.plugin.register_target("mytarget1")
    assert_equals(pyblish.plugin.current_target(), "mytarget1")

    assert len(pyblish.plugin.registered_targets()) == 2


@with_setup(lib.setup_empty, lib.teardown)
def test_register_target():
    """Registering and deregistering targets works fine"""
    pyblish.plugin.register_target("mytarget")
    assert "mytarget" in pyblish.plugin.registered_targets()
    pyblish.plugin.deregister_target("mytarget")
    assert "mytarget" not in pyblish.plugin.registered_targets()


@with_setup(lib.setup_empty, lib.teardown)
def test_data_dict():
    """.data is a pure dictionary"""

    context = pyblish.api.Context()
    instance = context.create_instance("MyInstance")
    assert isinstance(context.data, dict)
    assert isinstance(instance.data, dict)

    context.data["key"] = "value"
    assert context.data["key"] == "value"

    instance.data["key"] = "value"
    assert instance.data["key"] == "value"

    # Backwards compatibility
    assert context.data("key") == "value"
    assert instance.data("key") == "value"
    assert instance.data("name") == "MyInstance"
    # This returns (a copy of) the full dictionary
    assert context.data() == context.data


@with_setup(lib.setup_empty, lib.teardown)
def test_action():
    """Running an action is like running a plugin"""
    count = {"#": 0}

    class MyAction(pyblish.plugin.Action):
        def process(self, context):
            count["#"] += 1

    class MyPlugin(pyblish.plugin.Plugin):
        actions = [MyAction]

        def process(self, context):
            pass

    context = pyblish.api.Context()
    pyblish.plugin.process(
        plugin=MyPlugin,
        context=context,
        action=MyAction.id)

    assert count["#"] == 1


@with_setup(lib.setup_empty, lib.teardown)
def test_actions():
    class MyAction(pyblish.plugin.Action):
        def process(self, context):
            context.data["key"] = "value"

    context = pyblish.api.Context()
    pyblish.plugin.process(MyAction, context)
    assert "key" in context.data


@with_setup(lib.setup_empty, lib.teardown)
def test_action_error_checking():
    class MyActionValid(pyblish.plugin.Action):
        on = "all"

    class MyActionInvalid(pyblish.plugin.Action):
        on = "invalid"

    assert MyActionValid.__error__ is None
    assert MyActionInvalid.__error__


@with_setup(lib.setup_empty, lib.teardown)
def test_action_printing():
    class MyAction(pyblish.plugin.Action):
        pass

    print(MyAction())
    print(repr(MyAction()))

    assert str(MyAction()) == "MyAction"
    assert repr(MyAction()) == "pyblish.plugin.MyAction('MyAction')"


@with_setup(lib.setup_empty, lib.teardown)
def test_category_separator():
    assert issubclass(pyblish.plugin.Category("Test"),
                      pyblish.plugin.Action)
    assert issubclass(pyblish.plugin.Separator,
                      pyblish.plugin.Action)


def test_superclass_process_is_empty():
    """Superclass process() is empty"""
    def e():
        """Doc"""

    assert pyblish.api.Plugin.process.__code__.co_code == e.__code__.co_code
    assert pyblish.api.Plugin.repair.__code__.co_code == e.__code__.co_code


def test_plugin_source_path():
    """Plugins discovered carry a source path"""

    import sys
    plugin = pyblish.plugin.discover()[0]
    module = sys.modules[plugin.__module__]
    assert hasattr(module, "__file__")

    # Also works with inspect.getfile
    import inspect
    assert inspect.getfile(plugin) == module.__file__


@with_setup(lib.setup_empty, lib.teardown)
def test_register_callback():
    """Callback registration/deregistration works well"""

    def my_callback():
        pass

    def other_callback(data=None):
        pass

    pyblish.plugin.register_callback("mySignal", my_callback)

    msg = "Registering a callback failed"
    data = {"mySignal": [my_callback]}
    assert "mySignal" in pyblish.plugin.registered_callbacks() == data, msg

    pyblish.plugin.deregister_callback("mySignal", my_callback)

    assert_raises(
        ValueError,
        pyblish.plugin.deregister_callback,
        "mySignal", my_callback)

    assert_raises(
        KeyError,
        pyblish.plugin.deregister_callback,
        "notExist", my_callback)

    msg = "Deregistering a callback failed"
    data = {"mySignal": []}
    assert pyblish.plugin.registered_callbacks() == data, msg

    pyblish.plugin.register_callback("mySignal", my_callback)
    pyblish.plugin.register_callback("otherSignal", other_callback)
    pyblish.plugin.deregister_all_callbacks()

    msg = "Deregistering all callbacks failed"
    assert pyblish.plugin.registered_callbacks() == {}, msg


@with_setup(lib.setup_empty, lib.teardown)
def test_emit_signal_wrongly():
    """Exception from callback prints traceback"""

    def other_callback(an_argument=None):
        print("Ping from 'other_callback' with %s" % an_argument)

    pyblish.plugin.register_callback("otherSignal", other_callback)

    with lib.captured_stderr() as stderr:
        pyblish.lib.emit("otherSignal", not_an_argument="")
        output = stderr.getvalue().strip()
        print("Output: %s" % stderr.getvalue())
        assert output.startswith("Traceback")


@raises(ValueError)
@with_setup(lib.setup_empty, lib.teardown)
def test_registering_invalid_callback():
    """Can't register non-callables"""
    pyblish.plugin.register_callback("invalid", None)


@raises(KeyError)
def test_deregistering_nonexisting_callback():
    """Can't deregister a callback that doesn't exist"""
    pyblish.plugin.deregister_callback("invalid", lambda: "")


@raises(TypeError)
def test_register_noncallable_plugin():
    """Registered plug-ins must be callable"""
    pyblish.plugin.register_plugin("NotValid")


@raises(TypeError)
def test_register_old_plugin():
    """Can't register plug-ins incompatible with the version of Pyblish"""
    class MyPlugin(pyblish.plugin.Collector):
        requires = "pyblish==0"

    pyblish.plugin.register_plugin(MyPlugin)


@mock.patch("pyblish.plugin.__explicit_process")
def test_implicit_explicit_branching(func):
    """Explicit plug-ins are processed by the appropriate function"""

    # There are two mocks for this (see below); due to
    # @mock.patch.multiple being a dick.

    class Explicit(pyblish.plugin.ContextPlugin):
        pass

    pyblish.util.publish(plugins=[Explicit])
    assert func.call_count == 1, func.call_count


@mock.patch("pyblish.plugin.__implicit_process")
def test_implicit_branching(func):
    """Implicit plug-ins are processed by the appropriate function"""

    class Implicit(pyblish.plugin.Plugin):
        pass

    pyblish.util.publish(plugins=[Implicit])
    assert func.call_count == 1, func.call_count


def test_explicit_plugin():
    """ContextPlugin works as advertised"""

    count = {"#": 0}

    class Collector(pyblish.plugin.ContextPlugin):
        order = pyblish.plugin.CollectorOrder

        def process(self, context):
            self.log.info("Collecting from ContextPlugin")
            i = context.create_instance("MyInstance")
            i.data["family"] = "myFamily"
            count["#"] += 1

    class Validator(pyblish.plugin.InstancePlugin):
        order = pyblish.plugin.ValidatorOrder
        families = ["myFamily"]

        def process(self, instance):
            assert instance.data["name"] == "MyInstance", "fail"
            count["#"] += 10

    class ValidatorFailure(pyblish.plugin.InstancePlugin):
        order = pyblish.plugin.ValidatorOrder
        families = ["myFamily"]

        def process(self, instance):
            count["#"] += 100
            assert "notexist" in instance.data, "fail"

    class Extractor(pyblish.plugin.InstancePlugin):
        order = pyblish.plugin.ExtractorOrder
        families = ["myFamily"]

        def process(self, instance):
            count["#"] += 1000

    pyblish.util.publish(
        plugins=[
            Collector,
            Validator,
            ValidatorFailure,
            Extractor
        ]
    )

    assert count["#"] == 111, count


def test_context_plugin_wrong_arguments():
    """ContextPlugin doesn't take wrong arguments well"""

    count = {"#": 0}

    class Collector(pyblish.plugin.InstancePlugin):
        def process(self, context, instance):
            print("I won't run")
            count["#"] += 1

    pyblish.util.publish(plugins=[Collector])
    assert count["#"] == 0


def test_explicit_action():
    """Actions work with explicit plug-ins"""

    count = {"#": 0}

    class MyAction(pyblish.plugin.Action):
        def process(self, context):
            count["#"] += 1

    class MyPlugin(pyblish.plugin.ContextPlugin):
        actions = [MyAction]

        def process(self, context):
            pass

    context = pyblish.api.Context()
    pyblish.plugin.process(
        plugin=MyPlugin,
        context=context,
        action=MyAction.id)


def test_explicit_results():
    """Explicit plug-ins contain results"""

    class Collector(pyblish.plugin.ContextPlugin):
        order = pyblish.plugin.CollectorOrder

        def process(self, context):
            self.log.info("logged")

    context = pyblish.util.publish(plugins=[Collector])
    assert "results" in context.data

    result = context.data["results"][0]
    assert result["records"][0].msg == "logged"


def test_cooperative_collection():
    """Cooperative collection works

    A collector should be able to run such that the following
    collector "sees" the newly created instance so as to
    query and/or modify it.

    """

    count = {"#": 0}

    class CollectorA(pyblish.api.Collector):
        order = 0.0

        def process(self, context):
            context.create_instance("myInstance")
            count["#"] += 1

    class CollectorB(pyblish.api.Collector):
        order = 0.1

        def process(self, context):
            assert "myInstance" in [i.data["name"] for i in context]

            # This should run
            count["#"] += 10

    pyblish.api.register_plugin(CollectorA)
    pyblish.api.register_plugin(CollectorB)

    pyblish.util.publish()

    assert count["#"] == 11, count


def test_actions_and_explicit_plugins():
    """Actions work with explicit plug-ins"""

    count = {"#": 0}

    class MyAction(pyblish.api.Action):
        def process(self, context, plugin):
            count["#"] += 1
            raise Exception("Errored")

    class MyValidator(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder

        actions = [
            pyblish.api.Category("Scene"),
            MyAction
        ]

        def process(self, instance):
            count["#"] += 10

    context = pyblish.api.Context()
    result = pyblish.plugin.process(MyValidator,
                                    context,
                                    instance=None,
                                    action=MyAction.id)
    assert count["#"] == 1
    assert str(result["error"]) == "Errored", result


@with_setup(lib.setup_empty, lib.teardown)
def test_argumentless_plugin():
    """Plug-ins with neither instance nor context should still run"""
    count = {"#": 0}

    class MyPlugin(pyblish.api.Validator):
        def process(self):
            count["#"] += 1

    pyblish.api.register_plugin(MyPlugin)
    pyblish.util.publish()

    assert count["#"] == 1


@with_setup(lib.setup_empty, lib.teardown)
def test_argumentless_explitic_plugin():
    """Explicit plug-ins, without arguments, should fail"""
    class MyPlugin(pyblish.api.InstancePlugin):
        def process(self):
            pass

    raises(TypeError, pyblish.api.register_plugin, MyPlugin)

    class MyPlugin(pyblish.api.ContextPlugin):
        def process(self):
            pass

    raises(TypeError, pyblish.api.register_plugin, MyPlugin)


@with_setup(lib.setup_empty, lib.teardown)
def test_changes_to_registered_plugins_are_not_persistent():
    """Changes to registered plug-ins do not persist

    This is the expected behaviour of file-based plug-ins.

    """

    class MyPlugin(pyblish.api.ContextPlugin):
        active = False

    pyblish.api.register_plugin(MyPlugin)

    registered = pyblish.api.registered_plugins()[0]
    assert registered.id == MyPlugin.id
    assert registered.active is False

    registered.active = True
    assert registered.active is True

    registered = pyblish.api.registered_plugins()[0]
    assert registered.active is False


@with_setup(lib.setup_empty, lib.teardown)
def test_logging_solely_from_pyblish():
    """Only logging calls with self.log should be recorded."""

    class collect(pyblish.api.ContextPlugin):

        def process(self, context):
            import logging
            log = logging.getLogger("temp_logger")
            log.info("I should not be recorded!")

    context = pyblish.util.publish(plugins=[collect])
    for result in context.data["results"]:
        for record in result["records"]:
            assert record.name.startswith("pyblish")


@with_setup(lib.setup_empty, lib.teardown)
def test_running_for_all_targets():
    """Run for all targets when family is "default"."""

    count = {"#": 0}

    class plugin(pyblish.api.ContextPlugin):

        targets = ["default"]

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[plugin])

    assert count["#"] == 1, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_dont_run_non_matching_targets():
    """Don't run plugins that haven't got a target registered."""

    count = {"#": 0}

    class plugin(pyblish.api.ContextPlugin):

        targets = ["studio"]

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[plugin])

    assert count["#"] == 0, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_only_run_plugins_that_match_registered_targets():
    """Only run plugins that match the registered targets."""

    count = {"#": 0}

    class pluginStudio(pyblish.api.ContextPlugin):

        targets = ["studio"]

        def process(self, context):
            count["#"] += 1

    class pluginProject(pyblish.api.ContextPlugin):

        targets = ["project"]

        def process(self, context):
            count["#"] += 1

    pyblish.api.register_target("studio")
    pyblish.util.publish(plugins=[pluginStudio, pluginProject])

    assert count["#"] == 1, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_targets_and_exact_matching():
    """Run targets with exact matching."""

    count = {"#": 0}

    class pluginStudio(pyblish.api.ContextPlugin):

        targets = ["default", "studio"]
        match = pyblish.api.Exact

        def process(self, context):
            count["#"] += 1

    pyblish.api.register_target("studio")
    pyblish.util.publish(plugins=[pluginStudio])

    assert count["#"] == 1, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_targets_and_subset_matching():
    """Run targets with subset matching."""

    count = {"#": 0}

    class pluginStudio(pyblish.api.ContextPlugin):

        targets = ["studio"]
        match = pyblish.api.Subset

        def process(self, context):
            count["#"] += 1

    pyblish.api.register_target("studio")
    pyblish.util.publish(plugins=[pluginStudio])

    assert count["#"] == 1, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_targets_and_publishing():
    """Only run plugins with requested targets."""

    count = {"#": 0}

    class pluginA(pyblish.api.ContextPlugin):

        targets = ["customA"]

        def process(self, context):
            count["#"] += 1

    class pluginB(pyblish.api.ContextPlugin):

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[pluginA, pluginB], targets=["customA"])

    assert count["#"] == 1, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_targets_and_publishing_with_default():
    """Only run plugins with requested targets including default."""

    count = {"#": 0}

    class pluginA(pyblish.api.ContextPlugin):

        targets = ["customA"]

        def process(self, context):
            count["#"] += 1

    class pluginB(pyblish.api.ContextPlugin):

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(
        plugins=[pluginA, pluginB], targets=["default", "customA"]
    )

    assert count["#"] == 2, "count is {0}".format(count)


@with_setup(lib.setup_empty, lib.teardown)
def test_duplicate_plugin_names():
    logging.basicConfig(level=logging.DEBUG)

    pyblish.plugin.ALLOW_DUPLICATES = True

    plugins = []
    with lib.tempdir() as temp:
        plugin = (
            "from pyblish import api\nclass collectorA(api.ContextPlugin):"
            "\n    def process(self, context):\n        pass"
        )
        path = os.path.join(temp, "pluginA.py")
        with open(path, "w") as the_file:
            the_file.write(plugin)

        path = os.path.join(temp, "pluginA_copy.py")
        with open(path, "w") as the_file:
            the_file.write(plugin)

        plugins.extend(pyblish.api.discover(paths=[temp]))

    assert len(plugins) == 2, plugins
    
    # Restore state, for subsequent tests
    # NOTE: This assumes the test succeeds. If it fails, then
    # subsequent tests can fail because of it.
    pyblish.plugin.ALLOW_DUPLICATES = True




"""Shared processing logic"""

import os
import sys
import logging
import traceback

from . import _registered_test, _registered_gui, lib
from .plugin import (
    Validator,

    # Matchin algorithms
    Intersection,
    Subset,
    Exact,

    registered_targets
)

_algorithms = {
    Intersection: lambda a, b: set(a).intersection(b),
    Subset: lambda a, b: set(a).issubset(b),
    Exact: lambda a, b: set(a) == set(b)
}

log = logging.getLogger("pyblish.logic")


class TestFailed(Exception):
    def __init__(self, msg, vars):
        super(TestFailed, self).__init__(msg)
        self.vars = vars


def default_test(**vars):
    r"""Evaluate whether or not to continue processing

    The test determines whether or not to proceed from one
    plug-in to the next. The `vars` are updated everytime
    a plug-in is about to be processed with information about
    the upcoming plug-in.

    Returning any value means failure, whereas 0, False and None
    represents success. Similar to return/exit codes. You can provide
    a message along with a failure, such as specifying why the test
    failed. The message can then be used by process handlers,
    such as a GUI.

    You can provide your own test by registering it, see example below.

    Contents of `vars`:
        nextOrder (int): Order of next plugin
        ordersWithError (list): Orders at which an error has occured

    """

    offset = 0.5
    validation_order = Validator.order

    # If validation is done
    if vars["nextOrder"] >= validation_order + offset:
        for order in vars["ordersWithError"]:
            if lib.inrange(order,
                           base=validation_order,
                           offset=offset):
                return "failed validation"


def register_test(test):
    """Register test used to determine when to abort processing

    Arguments:
        test (callable): Called with argument `vars` and returns
            either True or False. True means to continue,
            False to abort.

    Example:
        >>> # Register custom test
        >>> def my_test(**vars):
        ...   return 1
        ...
        >>> register_test(my_test)
        >>>
        >>> # Run test
        >>> if my_test(order=1, ordersWithError=[]):
        ...   print("Test passed")
        Test passed
        >>>
        >>> # Restore default
        >>> deregister_test()

    """

    _registered_test["default"] = test


def registered_test():
    """Return the currently registered test"""
    return _registered_test["default"]


def deregister_test():
    """Restore default test"""
    register_test(default_test)


def register_gui(package):
    """Register a default GUI for Pyblish

    The argument `package` must refer to an available Python
    package with access to a `.show` member function taking no
    arguments. E.g.

    def show():
        pass

    This function is called whenever the default GUI
    is activated.

    Multiple GUIs:
        You may register more than one GUI, in which case each
        is tried in turn until a functioning match is found.

        For example, if both Pyblish QML and Pyblish Lite are
        registered, but Pyblish QML is not installed, then
        Pyblish Lite would appear as a "fallback".

    Arguments:
        package (str): Name of Python package with .show function.

    """

    if package not in _registered_gui:
        _registered_gui.append(package)


def registered_guis():
    """Return registered GUIs"""
    from_environment = os.environ.get(
        "PYBLISH_GUI", os.environ.get("PYBLISHGUI", "")
    )
    from_environment = list(gui for gui in from_environment.split(",") if gui)

    return _registered_gui[:] + from_environment


def deregister_gui(package):
    try:
        _registered_gui.remove(package)
    except ValueError:
        raise ValueError("\"%s\" has not been registered." % package)


def plugins_by_families(plugins, families):
    """Same as :func:`plugins_by_family` except it takes multiple families

    Arguments:
        plugins (list): List of plugins
        families (list): Families with which to compare against

    Returns:
        List of compatible plugins.

    """

    compatible = list()

    for plugin in plugins:

        if "*" in plugin.families:
            compatible.append(plugin)
            continue

        algorithm = _algorithms.get(plugin.match)

        assert algorithm, ("Plug-in did not provide "
                           "valid matching algorithm: %s" % plugin.match)

        if algorithm(plugin.families, families):
            compatible.append(plugin)

    return compatible


def plugins_by_family(plugins, family):
    """Convenience function to :func:`plugins_by_families`

    Arguments:
        plugins (list): List of plugins
        family (str): Family with which to compare against

    Returns:
        List of compatible plugins.

    """

    return plugins_by_families(plugins, [family])


def plugins_by_instance(plugins, instance):
    """Conveinence function for :func:`plugins_by_family`

    Arguments:
        plugins (list): Plug-ins to assess
        instance (Instance): Instance with which to compare against

    Returns:
        List of compatible plugins

    """

    family = instance.data.get("family")
    families = instance.data.get("families", [])

    return plugins_by_families(
        plugins, ([family] if family else []) + families)


def plugins_by_host(plugins, host):
    """Return compatible plugins `plugins` to host `host`

    Arguments:
        plugins (list): List of plugins
        host (str): Host with which compatible plugins are returned

    Returns:
        List of compatible plugins.

    """

    compatible = list()

    for plugin in plugins:
        # TODO(marcus): Expand to take partial wildcards e.g. "*Mesh"
        if any(x in getattr(plugin, "hosts", None) for x in (host, "*")):
            compatible.append(plugin)

    return compatible


def plugins_by_targets(plugins, targets):
    """Reutrn compatible plugins `plugins` to targets `targets`

    Arguments:
        plugins (list): List of plugins
        targets (list): List of targets with which to compare against

    Returns:
        List of compatible plugins.

    """

    compatible = list()

    for plugin in plugins:

        algorithm = _algorithms.get(plugin.match)

        assert algorithm, ("Plug-in did not provide "
                           "valid matching algorithm: %s" % plugin.match)

        if algorithm(plugin.targets, targets):
            compatible.append(plugin)

    return compatible


def instances_by_plugin(instances, plugin):
    """Return compatible instances `instances` to plugin `plugin`

    Return instances as they correspond to a plug-in, given
    an algorithm. The algorithm is determined by the `Plugin.match`

    When `match == Subset`, families of an instance must be a
    subset of families supported by a plug-in.

    Arguments:
        instances (list): List of instances
        plugin (Plugin): Plugin with which to compare against

    Returns:
        List of compatible instances

    Invariant:
        Order of remaining plug-ins must remain the same

    """

    algorithm = _algorithms.get(plugin.match)

    compatible = list()

    for instance in instances:

        if "*" in plugin.families:
            compatible.append(instance)
            continue

        assert algorithm, ("Plug-in did not provide "
                           "valid matching algorithm: %s" % plugin.match)

        family = instance.data.get("family")
        families = [family] if family else []
        families += instance.data.get("families", [])

        if algorithm(plugin.families, families):
            compatible.append(instance)

    return compatible


def _extract_traceback(exception):
    """Append traceback to `exception`

    This function safely extracts a traceback while being
    careful not to leak memory.

    Arguments:
        exception (Exception): Append traceback to here
            as "traceback" attribute.

    """

    exc_type = exc_value = exc_traceback = None

    try:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        exception.traceback = traceback.extract_tb(exc_traceback)[-1]

    except:
        pass

    finally:
        del(exc_type, exc_value, exc_traceback)


def Iterator(plugins, context, state=None):
    """Primary iterator

    This is the brains of publishing. It handles logic related
    to which plug-in to process with which Instance or Context,
    in addition to stopping when necessary.

    Arguments:
        plugins (list): Plug-ins to consider
        context (list): Instances to consider
        state (dict): Mutable state

    """

    test = registered_test()
    state = state or {
        "nextOrder": None,
        "ordersWithError": set()
    }

    # We'll add "default" target if no targets are registered. This happens
    # when running the Iterator directly without registering any targets.
    targets = registered_targets() or ["default"]

    plugins = plugins_by_targets(plugins, targets)

    for plugin in plugins:
        if not plugin.active:
            log.debug("%s was inactive, skipping.." % plugin)
            continue

        state["nextOrder"] = plugin.order

        message = test(**state)
        if message:
            raise StopIteration("Stopped due to %s" % message)

        instances = instances_by_plugin(context, plugin)
        if plugin.__instanceEnabled__:
            for instance in instances:
                if instance.data.get("publish") is False:
                    log.debug("%s was inactive, skipping.." % instance)
                    continue

                yield plugin, instance

        else:
            yield plugin, None



import os
import sys
import logging
import datetime
import warnings
import traceback
import functools

from . import _registered_callbacks
from .vendor import six


def inrange(number, base, offset=0.5):
    r"""Evaluate whether `number` is within `base` +- `offset`

    Lower bound is *included* whereas upper bound is *excluded*
    so as to allow for ranges to be stacked up against each other.
    For example, an offset of 0.5 and a base of 1 evenly stacks
    up against a base of 2 with identical offset.

    Arguments:
        number (float): Number to consider
        base (float): Center of range
        offset (float, optional): Amount of offset from base

    Usage:
        >>> inrange(0, base=1, offset=0.5)
        False
        >>> inrange(0.4, base=1, offset=0.5)
        False
        >>> inrange(1.4, base=1, offset=0.5)
        True
        >>> # Lower bound is included
        >>> inrange(0.5, base=1, offset=0.5)
        True
        >>> # Upper bound is excluded
        >>> inrange(1.5, base=1, offset=0.5)
        False

    """

    return (base - offset) <= number < (base + offset)


class MessageHandler(logging.Handler):
    def __init__(self, records, *args, **kwargs):
        # Not using super(), for compatibility with Python 2.6
        logging.Handler.__init__(self, *args, **kwargs)
        self.records = records

    def emit(self, record):
        if record.name.startswith("pyblish"):
            self.records.append(record)


def extract_traceback(exception):
    """Inject current traceback and store in exception"""
    exc_type, exc_value, exc_traceback = sys.exc_info()
    exception.traceback = traceback.extract_tb(exc_traceback)[-1]
    del(exc_type, exc_value, exc_traceback)


def time():
    """Return ISO formatted string representation of current UTC time."""
    return '%sZ' % datetime.datetime.utcnow().isoformat()


class ItemList(list):
    """List with keys

    Raises:
        KeyError is item is not in list

    Example:
        >>> Obj = type("Object", (object,), {})
        >>> obj = Obj()
        >>> obj.name = "Test"
        >>> l = ItemList(key="name")
        >>> l.append(obj)
        >>> l[0] == obj
        True
        >>> l["Test"] == obj
        True
        >>> try:
        ...   l["NotInList"]
        ... except KeyError:
        ...   print(True)
        True
        >>> obj == l.get("Test")
        True
        >>> l.get("NotInList") == None
        True

    """

    def __init__(self, key, object=list()):
        super(ItemList, self).__init__(object)
        self.key = key

    def __getitem__(self, index):
        if isinstance(index, int):
            return super(ItemList, self).__getitem__(index)

        for item in self:
            if getattr(item, self.key) == index:
                return item

        raise KeyError("%s not in list" % index)

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default


class classproperty(object):
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def log(cls):
    """Decorator for attaching a logger to the class `cls`

    Loggers inherit the syntax {module}.{submodule}

    Example
        >>> @log
        ... class MyClass(object):
        ...     pass
        >>>
        >>> myclass = MyClass()
        >>> myclass.log.info('Hello World')

    """

    module = cls.__module__
    name = cls.__name__

    # Package name appended, for filtering of LogRecord instances
    logname = "pyblish.%s.%s" % (module, name)
    cls.log = logging.getLogger(logname)

    # All messages are handled by root-logger
    cls.log.propagate = True

    return cls


def parse_environment_paths(paths):
    """Given a (semi-)colon separated string of paths, return a list

    Example:
        >>> import os
        >>> parse_environment_paths("path1" + os.pathsep + "path2")
        ['path1', 'path2']
        >>> parse_environment_paths("path1" + os.pathsep)
        ['path1', '']

    Arguments:
        paths (str): Colon or semi-colon (depending on platform)
            separated string of paths.

    Returns:
        list of paths as string.

    """

    paths_list = list()

    for path in paths.split(os.pathsep):
        paths_list.append(path)

    return paths_list


def get_formatter():
    """Return a default Pyblish formatter for logging

    Example:
        >>> import logging
        >>> log = logging.getLogger("myLogger")
        >>> handler = logging.StreamHandler()
        >>> handler.setFormatter(get_formatter())

    """

    formatter = logging.Formatter(
        '%(asctime)s - '
        '%(levelname)s - '
        '%(name)s - '
        '%(message)s',
        '%H:%M:%S')
    return formatter


def setup_log(root='pyblish', level=logging.DEBUG):
    """Setup a default logger for Pyblish

    Example:
        >>> log = setup_log()
        >>> log.info("Hello, World")

    """

    formatter = logging.Formatter("%(levelname)s - %(message)s")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    log = logging.getLogger(root)
    log.propagate = True
    log.handlers[:] = []
    log.addHandler(handler)

    log.setLevel(level)

    return log


def main_package_path():
    """Return path of main pyblish package"""
    lib_py_path = sys.modules[__name__].__file__
    package_path = os.path.dirname(lib_py_path)
    return package_path


def emit(signal, **kwargs):
    """Trigger registered callbacks

    Keyword arguments are passed from caller to callee.

    Arguments:
        signal (string): Name of signal emitted

    Example:
        >>> import sys
        >>> from . import plugin
        >>> plugin.register_callback(
        ...   "mysignal", lambda data: sys.stdout.write(str(data)))
        ...
        >>> emit("mysignal", data={"something": "cool"})
        {'something': 'cool'}

    """

    for callback in _registered_callbacks.get(signal, []):
        try:
            callback(**kwargs)
        except Exception:
            file = six.StringIO()
            traceback.print_exc(file=file)
            sys.stderr.write(file.getvalue())
            # Why the roundabout through StringIO?
            #
            # tests.lib.captured_stderr attempts to capture stderr
            # but doing so with plain print_exc() results in a type
            # error in Python 3. I'm not confident in Python 3 unicode
            # handling so there is likely a better way to solve this.
            #
            # TODO(marcus): Make it prettier


def deprecated(func):
    """Deprecation decorator

    Attach this to deprecated functions or methods.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper



import os

from . import lib

from pyblish import api, util
from nose.tools import (
    with_setup
)


def test_convenience_plugins_argument():
    """util._convenience() `plugins` argument works

    Issue: #286

    """

    count = {"#": 0}

    class PluginA(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            count["#"] += 1

    class PluginB(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            count["#"] += 10

    assert count["#"] == 0

    api.register_plugin(PluginA)
    util._convenience(0.5, plugins=[PluginB])

    assert count["#"] == 10, count


@with_setup(lib.setup, lib.teardown)
def test_convenience_functions():
    """convenience functions works as expected"""

    count = {"#": 0}

    class Collector(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            context.create_instance("MyInstance")
            count["#"] += 1

    class Validator(api.InstancePlugin):
        order = api.ValidatorOrder

        def process(self, instance):
            count["#"] += 10

    class Extractor(api.InstancePlugin):
        order = api.ExtractorOrder

        def process(self, instance):
            count["#"] += 100

    class Integrator(api.ContextPlugin):
        order = api.IntegratorOrder

        def process(self, instance):
            count["#"] += 1000

    class PostIntegrator(api.ContextPlugin):
        order = api.IntegratorOrder + 0.1

        def process(self, instance):
            count["#"] += 10000

    class NotCVEI(api.ContextPlugin):
        """This plug-in is too far away from Integration to qualify as CVEI"""
        order = api.IntegratorOrder + 2.0

        def process(self, instance):
            count["#"] += 100000

    assert count["#"] == 0

    for Plugin in (Collector,
                   Validator,
                   Extractor,
                   Integrator,
                   PostIntegrator,
                   NotCVEI):
        api.register_plugin(Plugin)

    context = util.collect()

    assert count["#"] == 1

    util.validate(context)

    assert count["#"] == 11

    util.extract(context)

    assert count["#"] == 111

    util.integrate(context)

    assert count["#"] == 11111


@with_setup(lib.setup, lib.teardown)
def test_multiple_instance_util_publish():
    """Multiple instances work with util.publish()

    This also ensures it operates correctly with an
    InstancePlugin collector.

    """

    count = {"#": 0}

    class MyContextCollector(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            context.create_instance("A")
            context.create_instance("B")
            count["#"] += 1

    class MyInstancePluginCollector(api.InstancePlugin):
        order = api.CollectorOrder + 0.1

        def process(self, instance):
            count["#"] += 1

    api.register_plugin(MyContextCollector)
    api.register_plugin(MyInstancePluginCollector)

    # Ensure it runs without errors
    util.publish()

    assert count["#"] == 3


@with_setup(lib.setup, lib.teardown)
def test_modify_context_during_CVEI():
    """Custom logic made possible via convenience members"""

    count = {"#": 0}

    class MyCollector(api.ContextPlugin):
        order = api.CollectorOrder

        def process(self, context):
            camera = context.create_instance("MyCamera")
            model = context.create_instance("MyModel")

            camera.data["family"] = "camera"
            model.data["family"] = "model"

            count["#"] += 1

    class MyValidator(api.InstancePlugin):
        order = api.ValidatorOrder

        def process(self, instance):
            count["#"] += 10

    api.register_plugin(MyCollector)
    api.register_plugin(MyValidator)

    context = api.Context()

    assert count["#"] == 0, count

    util.collect(context)

    assert count["#"] == 1, count

    context[:] = filter(lambda i: i.data["family"] == "camera", context)

    util.validate(context)

    # Only model remains
    assert count["#"] == 11, count

    # No further processing occurs.
    util.extract(context)
    util.integrate(context)

    assert count["#"] == 11, count


@with_setup(lib.setup, lib.teardown)
def test_environment_host_registration():
    """Host registration from PYBLISH_HOSTS works"""

    count = {"#": 0}
    hosts = ["test1", "test2"]

    # Test single hosts
    class SingleHostCollector(api.ContextPlugin):
        order = api.CollectorOrder
        host = hosts[0]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(SingleHostCollector)

    context = api.Context()

    os.environ["PYBLISH_HOSTS"] = "test1"
    util.collect(context)

    assert count["#"] == 1, count

    # Test multiple hosts
    api.deregister_all_plugins()

    class MultipleHostsCollector(api.ContextPlugin):
        order = api.CollectorOrder
        host = hosts

        def process(self, context):
            count["#"] += 10

    api.register_plugin(MultipleHostsCollector)

    context = api.Context()

    os.environ["PYBLISH_HOSTS"] = os.pathsep.join(hosts)
    util.collect(context)

    assert count["#"] == 11, count


@with_setup(lib.setup, lib.teardown)
def test_publishing_targets():
    """Publishing with targets works"""

    count = {"#": 0}

    class plugin(api.ContextPlugin):
        targets = ["custom"]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(plugin)

    util.publish(targets=["custom"])

    assert count["#"] == 1, count


@with_setup(lib.setup, lib.teardown)
def test_per_session_targets():
    """Register targets per session works"""

    util.publish(targets=["custom"])

    registered_targets = api.registered_targets()
    assert registered_targets == [], registered_targets


@with_setup(lib.setup, lib.teardown)
def test_publishing_collectors():
    """Running collectors with targets works"""

    count = {"#": 0}

    class plugin(api.ContextPlugin):
        order = api.CollectorOrder
        targets = ["custom"]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(plugin)

    util.collect(targets=["custom"])

    assert count["#"] == 1, count


@with_setup(lib.setup, lib.teardown)
def test_publishing_validators():
    """Running validators with targets works"""

    count = {"#": 0}

    class plugin(api.ContextPlugin):
        order = api.ValidatorOrder
        targets = ["custom"]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(plugin)

    util.validate(targets=["custom"])

    assert count["#"] == 1, count


@with_setup(lib.setup, lib.teardown)
def test_publishing_extractors():
    """Running extractors with targets works"""

    count = {"#": 0}

    class plugin(api.ContextPlugin):
        order = api.ExtractorOrder
        targets = ["custom"]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(plugin)

    util.extract(targets=["custom"])

    assert count["#"] == 1, count


@with_setup(lib.setup, lib.teardown)
def test_publishing_integrators():
    """Running integrators with targets works"""

    count = {"#": 0}

    class plugin(api.ContextPlugin):
        order = api.IntegratorOrder
        targets = ["custom"]

        def process(self, context):
            count["#"] += 1

    api.register_plugin(plugin)

    util.integrate(targets=["custom"])

    assert count["#"] == 1, count


@with_setup(lib.setup, lib.teardown)
def test_progress_existence():
    """Progress data member exists"""

    class plugin(api.ContextPlugin):
        pass

    api.register_plugin(plugin)

    result = next(util.publish_iter())

    assert "progress" in result, result


@with_setup(lib.setup, lib.teardown)
def test_publish_iter_increment_progress():
    """Publish iteration increments progress"""

    class pluginA(api.ContextPlugin):
        pass

    class pluginB(api.ContextPlugin):
        pass

    api.register_plugin(pluginA)
    api.register_plugin(pluginB)

    iterator = util.publish_iter()

    pluginA_progress = next(iterator)["progress"]
    pluginB_progress = next(iterator)["progress"]

    assert pluginA_progress < pluginB_progress



import os
import sys
import shutil
import tempfile

import pyblish
import pyblish.cli
import pyblish.api
from nose.tools import (
    with_setup,
    assert_equals,
)
from pyblish.vendor.click.testing import CliRunner
from . import lib

self = sys.modules[__name__]


def setup():
    self.tempdir = tempfile.mkdtemp()


def teardown():
    shutil.rmtree(self.tempdir)


def ctx():
    """Return current Click context"""
    return pyblish.cli._ctx


def context():
    """Return current context"""
    return ctx().obj["context"]


def test_visualise_environment_paths():
    """Visualising environment paths works well"""
    current_path = os.environ.get("PYBLISHPLUGINPATH")

    try:
        os.environ["PYBLISHPLUGINPATH"] = "/custom/path"

        runner = CliRunner()
        result = runner.invoke(pyblish.cli.main, ["--environment-paths"])

        assert result.output.startswith("/custom/path"), result.output

    finally:
        if current_path is not None:
            os.environ["PYBLISHPLUGINPATH"] = current_path


@with_setup(lib.setup_empty, lib.teardown)
def test_publishing():
    """Basic publishing works"""

    count = {"#": 0}

    class Collector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            self.log.warning("Running")
            count["#"] += 1
            context.create_instance("MyInstance")

    class MyValidator(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder

        def process(self, instance):
            count["#"] += 10
            assert instance.data["name"] == "MyInstance"
            count["#"] += 100

    pyblish.api.register_plugin(Collector)
    pyblish.api.register_plugin(MyValidator)

    runner = CliRunner()
    result = runner.invoke(pyblish.cli.main, ["publish"])
    print(result.output)

    assert count["#"] == 111, count


@with_setup(lib.setup, lib.teardown)
def test_environment_host_registration():
    """Host registration from PYBLISH_HOSTS works"""

    count = {"#": 0}
    hosts = ["test1", "test2"]

    # Test single hosts
    class SingleHostCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder
        host = hosts[0]

        def process(self, context):
            count["#"] += 1

    pyblish.api.register_plugin(SingleHostCollector)

    os.environ["PYBLISH_HOSTS"] = hosts[0]

    runner = CliRunner()
    result = runner.invoke(pyblish.cli.main, ["publish"])
    print(result.output)

    assert count["#"] == 1, count

    # Test multiple hosts
    pyblish.api.deregister_all_plugins()

    class MultipleHostsCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder
        host = hosts

        def process(self, context):
            count["#"] += 10

    pyblish.api.register_plugin(MultipleHostsCollector)

    os.environ["PYBLISH_HOSTS"] = os.pathsep.join(hosts)

    runner = CliRunner()
    result = runner.invoke(pyblish.cli.main, ["publish"])
    print(result.output)

    assert count["#"] == 11, count


@with_setup(lib.setup, lib.teardown)
def test_show_gui():
    """Showing GUI through cli works"""

    with tempfile.NamedTemporaryFile(dir=self.tempdir,
                                     delete=False,
                                     suffix=".py") as f:
        module_name = os.path.basename(f.name)[:-3]
        f.write(b"""\
def show():
    print("Mock GUI shown successfully")

if __name__ == '__main__':
    show()
""")

    pythonpath = os.pathsep.join([
        self.tempdir,
        os.environ.get("PYTHONPATH", "")
    ])

    print(module_name)

    runner = CliRunner()
    result = runner.invoke(
        pyblish.cli.main, ["gui", module_name],
        env={"PYTHONPATH": pythonpath}
    )

    assert_equals(result.output.splitlines()[-1].rstrip(),
                  "Mock GUI shown successfully")
    assert_equals(result.exit_code, 0)


@with_setup(lib.setup, lib.teardown)
def test_uses_gui_from_env():
    """Uses gui from environment var works"""

    with tempfile.NamedTemporaryFile(dir=self.tempdir,
                                     delete=False,
                                     suffix=".py") as f:
        module_name = os.path.basename(f.name)[:-3]
        f.write(b"""\
def show():
    print("Mock GUI shown successfully")

if __name__ == '__main__':
    show()
""")

    pythonpath = os.pathsep.join([
        self.tempdir,
        os.environ.get("PYTHONPATH", "")
    ])

    runner = CliRunner()
    result = runner.invoke(
        pyblish.cli.main, ["gui"],
        env={
            "PYTHONPATH": pythonpath,
            "PYBLISH_GUI": module_name
        }
    )

    assert_equals(result.output.splitlines()[-1].rstrip(),
                  "Mock GUI shown successfully")
    assert_equals(result.exit_code, 0)


@with_setup(lib.setup, lib.teardown)
def test_passing_data_to_gui():
    """Passing data to GUI works"""

    with tempfile.NamedTemporaryFile(dir=self.tempdir,
                                     delete=False,
                                     suffix=".py") as f:
        module_name = os.path.basename(f.name)[:-3]
        f.write(b"""\
from pyblish import util

def show():
    context = util.publish()
    print(context.data["passedFromTest"])

if __name__ == '__main__':
    show()
""")

    pythonpath = os.pathsep.join([
        self.tempdir,
        os.environ.get("PYTHONPATH", "")
    ])

    runner = CliRunner()
    result = runner.invoke(
        pyblish.cli.main, [
            "--data", "passedFromTest", "Data passed successfully",
            "gui", module_name
        ],
        env={"PYTHONPATH": pythonpath}
    )

    assert_equals(result.output.splitlines()[-1].rstrip(),
                  "Data passed successfully")
    assert_equals(result.exit_code, 0)
